TalcLock

Struct TalcLock 

pub struct TalcLock<R: RawMutex, S: Source, B: Binning> { /* private fields */ }
Wraps a mutex-locked Talc.

Example

use spinning_top::RawSpinlock;

let talc = talc::TalcLock::<RawSpinlock, _>::new(Manual);

Implementations

Source

impl<R: RawMutex, S: Source, B: Binning> TalcLock<R, S, B>

pub const fn new(source: S) -> Self

Create a new TalcLock.

pub fn lock(&self) -> MutexGuard<'_, R, Talc<S, B>>

Lock the mutex and access the inner Talc.

pub fn try_lock(&self) -> Option<MutexGuard<'_, R, Talc<S, B>>>

Try to lock the mutex and access the inner Talc.

pub fn get_mut(&mut self) -> &mut Talc<S, B>

Returns a mutable reference to the inner Talc.

This avoids locking, as having a mutable reference statically guarantees that self is not locked.

pub fn into_inner(self) -> Talc<S, B>

Retrieve the inner Talc.