pub struct TalcLock<R: RawMutex, S: Source, B: Binning> { /* private fields */ }Expand description
Implementations
Sourceimpl<R: RawMutex, S: Source, B: Binning> TalcLock<R, S, B>
impl<R: RawMutex, S: Source, B: Binning> TalcLock<R, S, B>
Sourcepub fn lock(&self) -> MutexGuard<'_, R, Talc<S, B>>
pub fn lock(&self) -> MutexGuard<'_, R, Talc<S, B>>
Lock the mutex and access the inner Talc.
Sourcepub fn try_lock(&self) -> Option<MutexGuard<'_, R, Talc<S, B>>>
pub fn try_lock(&self) -> Option<MutexGuard<'_, R, Talc<S, B>>>
Try to lock the mutex and access the inner Talc.
Sourcepub fn get_mut(&mut self) -> &mut Talc<S, B>
pub fn get_mut(&mut self) -> &mut Talc<S, B>
Returns a mutable reference to the inner Talc.
This avoids locking, as having a mutable reference statically
guarantees that self is not locked.
Sourcepub fn into_inner(self) -> Talc<S, B>
pub fn into_inner(self) -> Talc<S, B>
Retrieve the inner Talc.